/**
 * fBXNNX
 */

package hockey;

class Disk {
    int id, property;
    double x, y, vx, vy, r, k, e_d, e_t, e_p;
    boolean[] hit_d, hit_p;

    Disk(int id, int property, double r, double k, double e) {
        this.id = id;
        this.property = property;
        this.r = r;
        this.k = k;
        e_d = e * 1.0;
        e_t = e * 0.6;
        e_p = e * 0.2;
        hit_d = new boolean[id];
        hit_p = new boolean[2];
    }

    void set(double x, double y, double vx, double vy) {
        this.x = x;
        this.y = y;
        this.vx = vx;
        this.vy = vy;
    }

    void move(double dt) {
        x += vx*dt;
        y += vy*dt;
        vx += -k*vx*dt;
        vy += -k*vy*dt;
    }

    void bound(Disk d) {
        double ex, ey, l, d1, d2;
        l = Math.sqrt((x - d.x)*(x - d.x) + (y - d.y)*(y - d.y)) + 0.01;
        if(l < r + d.r) {
            if(!hit_d[d.id]) {
                ex = (x - d.x)/l;
                ey = (y - d.y)/l;
                d1 = ex*d.vx + ey*d.vy;
                d2 = ex*vx + ey*vy;
                d.vx = (d.vx - d1*ex) + ((1.0 - d.e_d)/2.0 * d1 + (1.0 + d.e_d)/2.0 * d2)*ex;
                d.vy = (d.vy - d1*ey) + ((1.0 - d.e_d)/2.0 * d1 + (1.0 + d.e_d)/2.0 * d2)*ey;
                vx = (vx - d2*ex) + ((1.0 + d.e_d)/2.0 * d1 + (1.0 - d.e_d)/2.0 * d2)*ex;
                vy = (vy - d2*ey) + ((1.0 + d.e_d)/2.0 * d1 + (1.0 - d.e_d)/2.0 * d2)*ey;
                hit_d[d.id] = true;
            }
        } else if(hit_d[d.id]) {
            hit_d[d.id] = false;
        }
    }
}