/**
 * ͂ρ` GAzbP[
 * Created on 2005/02/20
 * @author Takuya (http://clustera.skr.jp/java/)
 * @version 1.5
 */

package hockey;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.applet.Applet;
import library.graphics.*;

public class Hockey extends Applet implements Runnable, MouseListener, MouseMotionListener, KeyListener {
    private final static int SWIDTH = 400, SHEIGHT = 380, MAX = 5;
    private final static double TWIDTH = 240, THEIGHT = 360, R_PAD = 10, R_DISK = 9;

    private boolean saving = false, wireframe = false;
    private int mode, level, score_p, score_e, count, div;
    private double dt, oldx1, oldy1, oldx2, oldy2, newx, newy;
    private long current_time, play_time, calc_time, old_time;
    private long[] goal_time = new long[2];
    private long[] effect_time = new long[4];

    private Thread th;
    private Image buff;
    private Graphics buff_g;

    private Font font16p = new Font("SansSerif", Font.PLAIN, 16);
    private Font font24p = new Font("SansSerif", Font.PLAIN, 24);
    private Font font16b = new Font("SansSerif", Font.BOLD, 16);
    private Font font32b = new Font("SansSerif", Font.BOLD, 32);
    private Font font48b = new Font("SansSerif", Font.BOLD, 48);

    private Table table = new Table(TWIDTH, THEIGHT);
    private Pad pad_p = new Pad(R_PAD, TWIDTH, THEIGHT);
    private Enemy pad_e = new Enemy(R_PAD, TWIDTH, THEIGHT);
    private Disk[] disk = new Disk[MAX];

    private ZRenderer renderer = new ZRenderer(this);
    private Light light = new Light();

    private Model[] mtable = new Model[5];
    private Model mpad1 = new Model();
    private Model mpad2 = new Model();
    private Model mdisk = new Model();
    private Model backscreen = new Model();

    private Material white =       new Material(Array.make(1.0, 1.0, 1.0));
    private Material blue =        new Material(Array.make(0.5, 0.5, 1.0));
    private Material redpurple =   new Material(Array.make(0.9, 0.3, 0.5));
    private Material bluepurple =  new Material(Array.make(0.5, 0.3, 0.9));
    private Material lightgray =   new Material(Array.make(0.9, 0.9, 0.9));
    private Material lightgreen =  new Material(Array.make(0.9, 1.0, 0.9));
    private Material lightyellow = new Material(Array.make(1.0, 1.0, 0.9));
    private Material[] sky = new Material[4];

    public void init() {
        calc_time = 1000;
        dt = 1.0/600.0;
        old_time = System.currentTimeMillis();

        for(int i=0; i<MAX; i++)
            disk[i] = new Disk(i, (int)(Math.random()*10.0), R_DISK, 0.15, 0.75);

        renderer.setViewport(0, 80, SWIDTH, SHEIGHT-80);
        renderer.initProjection();
        renderer.setPerspective(0.75, SWIDTH/(SHEIGHT-80), 100.0, 700.0);

        for(int i=0; i<5; i++) mtable[i] = new Model();
        mtable[1].setPlane(TWIDTH+2, 6);
        mtable[2].setBox(TWIDTH*0.3, 12, 10);
        mtable[3].setBox(TWIDTH*0.4, 6, 8);
        mtable[4].setBox(12, THEIGHT+24, 12);
        backscreen.setPlane(SWIDTH*2, SWIDTH*2);

        setModel(saving);

        for(int i=0; i<4;i++) sky[i] = new Material(Array.make(0.0, 0.0, 0.0));

        prepare(1);

        if(th == null) {
            th = new Thread(this);
            th.start();
        }
        addMouseListener(this);
        addMouseMotionListener(this);
        addKeyListener(this);

        setSize(SWIDTH, SHEIGHT);
    }

    private void control() {
        play_time += calc_time;

        if(current_time - effect_time[0] < 20000)
            pad_p.r = R_PAD * 1.5;
        else if(pad_p.r != R_PAD)
            pad_p.r = R_PAD;
        if(current_time - effect_time[1] < 20000)
            pad_e.r = R_PAD * 1.5;
        else if(pad_e.r != R_PAD)
            pad_e.r = R_PAD;

        if((current_time - effect_time[2] < 20000) || (current_time - effect_time[3] < 20000))
            for(int i=0; i<level; i++) disk[i].r = R_DISK * 0.5;
        else if(disk[0].r != R_DISK)
            for(int i=0; i<level; i++) disk[i].r = R_DISK;

        oldx1 = pad_p.x;
        oldy1 = pad_p.y;
        oldx2 = oldx1;
        oldy2 = oldy1;
        for(int k=1; k<=div; k++) {
            pad_p.x = (newx*k + oldx1*(div - k))/div;
            pad_p.y = (newy*k + oldy1*(div - k))/div;
            table.bound(pad_p, 0);
            pad_p.vx = (pad_p.x - oldx2)/dt;
            pad_p.vy = (pad_p.y - oldy2)/dt;
            oldx2 = pad_p.x;
            oldy2 = pad_p.y;

            pad_e.move(disk, dt, level);
            table.bound(pad_e, 1);

            for(int i=0; i<level; i++) {
                disk[i].move(dt);
                for(int j=0; j<i; j++) disk[i].bound(disk[j]);
                pad_p.bound(disk[i], 0);
                pad_e.bound(disk[i], 1);

                if((disk[i].x > -TWIDTH*0.2 + disk[i].r) && (disk[i].x < TWIDTH*0.2 - disk[i].r)) {
                    if((disk[i].y < -THEIGHT/2) || (disk[i].y > THEIGHT/2)) {
                        if(disk[i].y < 0) {
                            score_p++;
                            goal_time[0] = current_time;
                            if(disk[i].property == 0)
                                effect_time[0] = current_time;
                            else if(disk[i].property == 1) {
                                score_p++;
                                effect_time[2] = current_time;
                            }
                            if(score_p >= (level/5 + 1)*3) mode = 3;
                        } else {
                            score_e++;
                            goal_time[1] = current_time;
                            if(disk[i].property == 0)
                                effect_time[1] = current_time;
                            else if(disk[i].property == 1) {
                                score_e++;
                                effect_time[3] = current_time;
                            }
                            if(score_e >= (level/5 + 1)*3) mode = 3;
                        }
                        disk[i].property = (int)(Math.random()*10.0);
                        double r = Math.random();
                        double cos = 50.0*Math.cos(2*Math.PI*r);
                        double sin = 50.0*Math.sin(2*Math.PI*r);
                        disk[i].set(cos, sin, cos*2, sin*2);
                    }
                } else {
                    table.bound(disk[i]);
                }
            }
        }
    }

    private void display() {
        renderer.initModelview();
        renderer.setLookAt(1.5*TWIDTH, 0.7 + 0.4*pad_p.y/THEIGHT, Math.PI/2 - 0.3*pad_p.x/TWIDTH);
        renderer.setTranslate(0, -TWIDTH/4, 0);
        renderer.clear();

        light.setSpecular(Array.make(0.75, 0.75, 0.75));
        renderer.setLight(light);

        renderer.pushMatrix();
            renderer.setMaterial(blue);
            renderer.drawModel(mtable[0]);
            renderer.setTranslate(0, 0, 1);
            renderer.setMaterial(white);
            renderer.drawModel(mtable[1]);
        renderer.popMatrix();

        light.setSpecular(Array.make(0.25, 0.25, 0.25));
        renderer.setLight(light);

        renderer.setTranslate(0, 0, 5);
        renderer.setMaterial(lightgray);
        renderer.pushMatrix();
            renderer.setTranslate(-TWIDTH*0.35, THEIGHT/2 + 6, 0);
            renderer.drawModel(mtable[2]);
            renderer.setTranslate(TWIDTH*0.7, 0, 0);
            renderer.drawModel(mtable[2]);
            renderer.setTranslate(0,-THEIGHT - 12, 0);
            renderer.drawModel(mtable[2]);
            renderer.setTranslate(-TWIDTH*0.7, 0, 0);
            renderer.drawModel(mtable[2]);
        renderer.popMatrix();
        if(!saving) {
            renderer.pushMatrix();
                renderer.setTranslate(0, -THEIGHT/2 - 9, -1);
                renderer.drawModel(mtable[3]);
                renderer.setTranslate(0, THEIGHT + 18, 0);
                renderer.drawModel(mtable[3]);
                renderer.setTranslate(-TWIDTH/2 - 6, -THEIGHT/2 - 9, 2);
                renderer.drawModel(mtable[4]);
                renderer.setTranslate(TWIDTH + 12, 0, 0);
                renderer.drawModel(mtable[4]);
            renderer.popMatrix();
        }
        renderer.setTranslate(0, 0, -5);

        renderer.setMaterial(lightyellow);
        if(mode != 1)
            for(int i=0; i<level; i++) {
                renderer.pushMatrix();
                    renderer.setTranslate(disk[i].x, disk[i].y, R_DISK/4);
                    renderer.setScale(disk[0].r/R_DISK, disk[0].r/R_DISK, 1.0);
                    renderer.setRotateX(Math.PI/2);
                    if(disk[i].property == 0) {
                        renderer.setMaterial(redpurple);
                        renderer.drawModel(mdisk);
                        renderer.setMaterial(lightyellow);
                    } else if(disk[i].property == 1) {
                        renderer.setMaterial(blue);
                        renderer.drawModel(mdisk);
                        renderer.setMaterial(lightyellow);
                    } else {
                        renderer.drawModel(mdisk);
                    }
                renderer.popMatrix();
            }

        renderer.setMaterial(lightgreen);
        renderer.pushMatrix();
            renderer.setTranslate(pad_p.x, pad_p.y, pad_p.r/4);
            renderer.setScale(pad_p.r/R_DISK, pad_p.r/R_DISK, pad_p.r/R_DISK);
            renderer.setRotateX(Math.PI/2);
            renderer.drawModel(mpad1);
            renderer.setTranslate(0, R_PAD/4, 0);
            renderer.drawModel(mpad2);
        renderer.popMatrix();
        renderer.pushMatrix();
            renderer.setTranslate(pad_e.x, pad_e.y, pad_e.r/4);
            renderer.setScale(pad_e.r/R_DISK, pad_e.r/R_DISK, pad_e.r/R_DISK);
            renderer.setRotateX(Math.PI/2);
            renderer.drawModel(mpad1);
            renderer.setTranslate(0, R_PAD/4, 0);
            renderer.drawModel(mpad2);
        renderer.popMatrix();

        if(!saving) {
            renderer.pushMatrix();
                renderer.initModelview();
                renderer.setTranslate(0, 0, -600.0);
                renderer.drawModel(backscreen, sky);
            renderer.popMatrix();
        }

        renderer.draw(buff_g);
    }

    private void score() {
        buff_g.setColor(Color.black);
        buff_g.fillRect(0, 0, SWIDTH, 80);

        buff_g.setFont(font16p);
        buff_g.setColor(Color.white);
        buff_g.drawString(Double.toString((int)(2000000.0/calc_time)/100.0) + " fps", 5, 15);
        buff_g.setColor(Color.green);
        buff_g.drawString("TIME : " + Long.toString(play_time/20000), SWIDTH/2 + 100, 15);

        buff_g.setFont(font16b);
        buff_g.drawString("PLAYER", SWIDTH/2-50 - 59, 78);
        buff_g.drawString("ENEMY", SWIDTH/2+50, 78);

        buff_g.setColor(Color.red);
        if(current_time - effect_time[0] < 20000)
            buff_g.drawString("RED " + Long.toString((20000 - current_time + effect_time[0])/1000 + 1), SWIDTH/2 - 190, 45);
        if(current_time - effect_time[1] < 20000)
            buff_g.drawString("RED " + Long.toString((20000 - current_time + effect_time[1])/1000 + 1), SWIDTH/2 + 120, 45);
        buff_g.setColor(Color.blue);
        if(current_time - effect_time[2] < 20000)
            buff_g.drawString("BLUE " + Long.toString((20000 - current_time + effect_time[2])/1000 + 1), SWIDTH/2 - 190, 60);
        if(current_time - effect_time[3] < 20000)
            buff_g.drawString("BLUE " + Long.toString((20000 - current_time + effect_time[3])/1000 + 1), SWIDTH/2 + 120, 60);

        buff_g.setFont(font24p);
        buff_g.setColor(Color.pink);
        buff_g.drawString("LEVEL : " + Integer.toString(level), SWIDTH/2 - 55, 20);

        buff_g.setFont(font48b);
        buff_g.setColor(Color.orange);
        buff_g.drawString("-", SWIDTH/2 - 8, 60);
        buff_g.drawString(Integer.toString(score_p), SWIDTH/2-50 - 27*(score_p/10 + 1), 60);
        buff_g.drawString(Integer.toString(score_e), SWIDTH/2+50, 60);

        switch(mode) {
            case 1:
                if(current_time%500 > 200) {
                    buff_g.setFont(font32b);
                    buff_g.setColor(Color.white);
                    buff_g.drawString("READY ?", SWIDTH/2 - 65, SHEIGHT/2);
                }
                break;
            case 2:
                if((current_time - goal_time[0] < 1000) && (current_time%250 > 100)) {
                    buff_g.setColor(Color.black);
                    buff_g.drawString(Integer.toString(score_p), SWIDTH/2-50 - 27*(score_p/10 + 1), 60);
                    if(score_p == (level/5 + 1)*3 - 1) {
                        buff_g.setFont(font32b);
                        buff_g.setColor(Color.red);
                        buff_g.drawString("MatchPoint !", SWIDTH/2 - 90, SHEIGHT/2);
                    }
                }
                if((current_time - goal_time[1] < 1000) && (current_time%250 > 100)) {
                    buff_g.setColor(Color.black);
                    buff_g.drawString(Integer.toString(score_e), SWIDTH/2+50, 60);
                    if(score_e == (level/5 + 1)*3 - 1) {
                        buff_g.setFont(font32b);
                        buff_g.setColor(Color.blue);
                        buff_g.drawString("MatchPoint !", SWIDTH/2 - 90, SHEIGHT/2);
                    }
                }
                break;
            case 3:
                if(current_time%500 > 200) {
                    buff_g.setFont(font32b);
                    if(score_p >= (level/5 + 1)*3) {
                        buff_g.setColor(Color.red);
                        buff_g.drawString("YOU WIN", SWIDTH/2 - 65, SHEIGHT/2);
                    } else {
                        buff_g.setColor(Color.blue);
                        buff_g.drawString("YOU LOSE", SWIDTH/2 - 75, SHEIGHT/2);
                    }
                }
                break;
            case 4:
                buff_g.setColor(Color.yellow);
                buff_g.setFont(font24p);
                buff_g.drawString("CLEAR TIME : " + Long.toString(play_time/20000), SWIDTH/2 - 100, SHEIGHT/2 + 40);
                if(current_time%500 > 200) {
                    buff_g.setFont(font32b);
                    buff_g.drawString("Congratulations !", SWIDTH/2 - 130, SHEIGHT/2);
                }
                break;
        }
    }

    private void prepare(int level) {
        this.level = level;
        mode = 1;
        score_p = 0;
        score_e = 0;
        if(level == 1) play_time = 0;
        for(int i=0; i<4; i++) effect_time[i] = current_time - 20000;

        pad_p.set(0.0, THEIGHT/2 - pad_p.r, 0.0, 0.0);
        pad_e.set(0.0,-THEIGHT/2 + pad_e.r, 0.0, 0.0);

        for(int i=0; i<level; i++) {
            double cos = 50.0*Math.cos(2*Math.PI*i/level + Math.PI/2);
            double sin = 50.0*Math.sin(2*Math.PI*i/level + Math.PI/2);
            disk[i].set(cos, sin, cos*2, sin*2);
        }

        light.setPosition(Array.make((MAX+1 - 2*level)*10.0, -200.0, (2*MAX - level)*40.0));
        light.setDiffuse(Array.make(1.0, 1.0 - 0.3*level/MAX, 1.0 - 0.6*level/MAX));

        sky[0].setAmbient(Array.make(2.5 + 0.4*level, 4.0 + 0.2*level, 5.0 - 0.8*level));
        sky[1].setAmbient(Array.make(2.5 + 0.4*level, 4.0 + 0.2*level, 5.0 - 0.8*level));
        sky[2].setAmbient(Array.make(0.8*level, 1.0 + 0.12*level, 5.0 - 0.8*level));
        sky[3].setAmbient(Array.make(0.8*level, 1.0 + 0.12*level, 5.0 - 0.8*level));
        renderer.setClearColor(Array.make(0.16*level, 0.2 + 0.024*level, 1.0 - 0.16*level));
    }

    public void paint(Graphics g) {
        if(buff == null) {
            buff = createImage(SWIDTH, SHEIGHT);
            buff_g = buff.getGraphics();
        }

        current_time = System.currentTimeMillis();
        if(count++ == 20) {
            calc_time = current_time - old_time;
            div = (int)(0.03*calc_time);
            old_time = current_time;
            count = 0;
        }

        if(mode == 2) control();
        display();
        score();

        g.drawImage(buff, 0, 0, this);
    }

    public void update(Graphics g) {
        paint(g);
    }

    public void run() {
        while(true) {
            try {
                Thread.sleep(10);
                repaint();
            } catch(InterruptedException e) {
            }
        }
    }

    private void setModel(boolean saving) {
        if(saving) {
            mtable[0].setPlane(1, TWIDTH + 2, THEIGHT + 14);
            mpad1.setCylinder(8, R_PAD, R_PAD/2);
            mpad2.setCylinder(4, R_PAD/2, R_PAD);
            mdisk.setCylinder(8, R_DISK, R_DISK/2);
        } else {
            mtable[0].setPlane(8, TWIDTH + 2, THEIGHT + 14);
            mpad1.setCylinder(12, R_PAD, R_PAD/2);
            mpad2.setCylinder(6, R_PAD/2, R_PAD);
            mdisk.setCylinder(12, R_DISK, R_DISK/2);
        }
    }

    public void mouseClicked(MouseEvent e) {
        switch(mode) {
            case 1:
                mode = 2;
                break;
            case 3:
                if(score_p >= (level/5 + 1)*3) {
                    if(level < MAX) prepare(level + 1);
                    else mode = 4;
                } else {
                    prepare(level);
                }
                break;
            case 4:
                prepare(1);
                break;
        }
    }
    public void mouseEntered(MouseEvent e) {
    }
    public void mouseExited(MouseEvent e) {
    }
    public void mousePressed(MouseEvent e) {
    }
    public void mouseReleased(MouseEvent e) {
    }
    public void mouseDragged(MouseEvent e) {
    }
    public void mouseMoved(MouseEvent e) {
        newx = (SWIDTH/2 - e.getX()) * 0.7;
        newy = (e.getY() - SHEIGHT/2) * 1.2;
    }

    public void keyPressed(KeyEvent e) {
        if(e.getKeyCode() == e.VK_SPACE) {
            saving = !saving;
            setModel(saving);
        }
        if(e.getKeyCode() == e.VK_ENTER) {
            if(wireframe) renderer.disable(renderer.MODEL_WIREFRAME);
            else renderer.enable(renderer.MODEL_WIREFRAME);
            wireframe = !wireframe;
        }
    }
    public void keyReleased(KeyEvent e) {
    }
    public void keyTyped(KeyEvent e) {
    }
}