/**
 * _[iZ-BufferjNX
 * Created on 2004/09/24
 * @author Takuya (http://clustera.skr.jp/java/)
 * @version 1.3
 */

package library.graphics;

import java.awt.*;
import java.awt.image.*;
import java.util.Stack;

public final class ZRenderer {
    private int[] pbuf, zbuf;
    private int x, y, width, height, clearcolor;
    private Component c;
    private MemoryImageSource mis;
    private Image image;

    private double[][] matViewport, matProjection, matModelview, matTransform;
    private Stack st;

    public final static int MODEL_WIREFRAME = 0, MODEL_SURFACE = 1, SHADING_FLAT = 2, SHADING_SMOOTH = 3;
    public final static int SPEC_FACE = 4, SPEC_VERTEX = 5, CULLFACE = 6;
    private boolean mode_surface, mode_smooth, mode_vertex, mode_cull;

    private double[] color;
    private Material material;
    private Light light;

    public ZRenderer(Component component) {
        c = component;

        clearcolor = 0xFFFFFFFF;
        st = new Stack();

        mode_surface = true;
        mode_smooth = true;
        mode_vertex = true;
        mode_cull = true;

        color = new double[3];
        material = new Material();
        light = new Light();
    }

    //sNZobt@AfvXobt@̏
    public void clear() {
        for(int i=0; i < width*height; i++) {
            pbuf[i] = clearcolor;
            zbuf[i] = 0x80000000;
        }
    }

    //sNZobt@F̐ݒ
    public void setClearColor(double[] color) {
        clearcolor = (255<<24) | ((int)(color[0]*255)<<16) | ((int)(color[1]*255)<<8) | (int)(color[2]*255);
    }

    //C[W̏o
    public void draw(Graphics g) {
        mis.newPixels();
        g.drawImage(image, x, y, c);
    }

    //r[|[g̐ݒ
    public void setViewport(int x, int y, int width, int height) {
        double[][] m = {
            {width/2.0,         0.0, 0.0,  width/2.0},
            {      0.0, -height/2.0, 0.0, height/2.0},
            {      0.0,         0.0, 1.0,        0.0},
            {      0.0,         0.0, 0.0,        1.0}
        };
        matViewport = m;

        this.x = x;
        this.y = y;
        this.width = width;
        this.height = height;

        pbuf = new int[width*height];
        zbuf = new int[width*height];
        mis = new MemoryImageSource(width, height, pbuf, 0, width);
        mis.setAnimated(true);
        image = c.createImage(mis);
    }

    //ˉes̏
    public void initProjection() {
        matProjection = Matrix.identity();
    }

    //ˉe̐ݒ
    public void setOrtho(double l, double r, double b, double t, double n, double f) {
        double[][] m = {
            {2.0/(r - l),         0.0,         0.0, -(r + l)/(r - l)},
            {        0.0, 2.0/(t - b),         0.0, -(t + b)/(t - b)},
            {        0.0,         0.0, 2.0/(n - f), -(n + f)/(n - f)},
            {        0.0,         0.0,         0.0,              1.0}
        };
        Matrix.multiply(matProjection, m);
    }

    //ˉe̐ݒ
    public void setFrustum(double l, double r, double b, double t, double n, double f) {
        double[][] m = {
            {2.0*n/(r - l),           0.0,  (r + l)/(r - l),              0.0},
            {          0.0, 2.0*n/(t - b),  (t + b)/(t - b),              0.0},
            {          0.0,           0.0, -(n + f)/(n - f), -2.0*n*f/(n - f)},
            {          0.0,           0.0,             -1.0,              0.0}
        };
        Matrix.multiply(matProjection, m);
    }
    public void setPerspective(double fovy, double aspect, double n, double f) {
        double t = n*Math.tan(fovy/2);
        double r = aspect*t;
        setFrustum(-r, r, -t, t, n, f);
    }

    //fr[s̏
    public void initModelview() {
        matModelview = Matrix.identity();
    }

    //_AQƓ_̐ݒ
    public void setLookAt(double[] eye, double[] center, double[] up) {
        double[] ec = Vector.normalize(Vector.sub(eye, center));
        double[] eb = Vector.normalize(up);
        double[] ea = Vector.cross(eb, ec);
        double[][] m = {
            {ea[0], ea[1], ea[2], -Vector.dot(ea, eye)},
            {eb[0], eb[1], eb[2], -Vector.dot(eb, eye)},
            {ec[0], ec[1], ec[2], -Vector.dot(ec, eye)},
            {  0.0,   0.0,   0.0,                  1.0}
        };
        Matrix.multiply(matModelview, m);
    }
    public void setLookAt(double distance, double theta, double phi) {
        setTranslate(0.0, 0.0, -distance);
        setRotateZ(-Math.PI/2);
        setRotateY(-theta);
        setRotateZ(-phi);
    }

    //]̐ݒ
    public void setRotateX(double angle) {
        Matrix.rotateX(matModelview, angle);
    }
    public void setRotateY(double angle) {
        Matrix.rotateY(matModelview, angle);
    }
    public void setRotateZ(double angle) {
        Matrix.rotateZ(matModelview, angle);
    }

    //gk̐ݒ
    public void setScale(double sx, double sy, double sz) {
        Matrix.scale(matModelview, sx, sy, sz);
    }

    //sړ̐ݒ
    public void setTranslate(double dx, double dy, double dz) {
        Matrix.translate(matModelview, dx, dy, dz);
    }

    //CӍs̃fr[sւ̏Z
    public void multMatrix(double[][] m) {
        Matrix.multiply(matModelview, m);
    }

    //fr[s̃X^bNւ̕ۑ
    public void pushMatrix() {
        st.push(Array.copy(matModelview));
    }

    //X^bNos̃fr[sւ̐ݒ
    public void popMatrix() {
        matModelview = (double[][])st.pop();
    }

    //e탂[h̗L
    public void enable(int mode) {
        switch(mode) {
            case MODEL_WIREFRAME: mode_surface = false; break;
            case MODEL_SURFACE:   mode_surface = true;  break;
            case SHADING_FLAT:    mode_smooth =  false; break;
            case SHADING_SMOOTH:  mode_smooth =  true;  break;
            case SPEC_FACE:       mode_vertex =  false; break;
            case SPEC_VERTEX:     mode_vertex =  true;  break;
            case CULLFACE:        mode_cull =    true;  break;
        }
    }

    //e탂[h̖
    public void disable(int mode) {
        switch(mode) {
            case MODEL_WIREFRAME: mode_surface = true;  break;
            case MODEL_SURFACE:   mode_surface = false; break;
            case SHADING_FLAT:    mode_smooth =  true;  break;
            case SHADING_SMOOTH:  mode_smooth =  false; break;
            case SPEC_FACE:       mode_vertex =  true;  break;
            case SPEC_VERTEX:     mode_vertex =  false; break;
            case CULLFACE:        mode_cull =    false; break;
        }
    }

    //`F̐ݒ
    public void setColor(double[] color) {
        this.color = Array.copy(color);
    }

    //ގ̐ݒ
    public void setMaterial(Material material) {
        this.material.ambient = Array.copy(material.ambient);
        this.material.diffuse = Array.copy(material.diffuse);
        this.material.specular = Array.copy(material.specular);
        this.material.shininess = material.shininess;
    }

    //̐ݒ
    public void setLight(Light light) {
        this.light.position = Matrix.transform3(matModelview, light.position);
        this.light.ambient = Array.copy(light.ambient);
        this.light.diffuse = Array.copy(light.diffuse);
        this.light.specular = Array.copy(light.specular);
    }

    //f̕`
    public void drawModel(Model model) {
        if(mode_surface) {
            if(mode_smooth) {
                //X[XVF[fBO
                int[][] tlv = new int[model.vertex.length][];
                for(int i=0; i < model.vertex.length; i++)
                    tlv[i] = transform_and_lighting(model.vertex[i], model.normal_vertex[i], material);

                for(int i=0; i < model.face.length; i++) {
                    int[][] tlf = {
                        tlv[model.face[i][0]], tlv[model.face[i][1]], tlv[model.face[i][2]]
                    };
                    if((tlf[0][6] + tlf[1][6] + tlf[2][6]) < 3)
                        drawSmoothTriangle(tlf);
                }
            } else {
                //tbgVF[fBO
                matTransform = Array.copy(matProjection);
                Matrix.multiply(matTransform, matModelview);

                int[][] tv = new int[model.vertex.length][];
                for(int i=0; i < model.vertex.length; i++)
                    tv[i] = transform(model.vertex[i]);

                for(int i=0; i < model.face.length; i++) {
                    int[][] tlf = {
                        tv[model.face[i][0]], tv[model.face[i][1]], tv[model.face[i][2]],
                        lighting(model.vertex[model.face[i][0]], model.vertex[model.face[i][1]],
                                model.vertex[model.face[i][2]], model.normal_face[i], material)
                    };
                    if((tlf[0][3] + tlf[1][3]+ tlf[2][3] + tlf[3][3]) < 3)
                        drawFlatTriangle(tlf);
                }
            }
        } else {
            //C[t[
            matTransform = Array.copy(matProjection);
            Matrix.multiply(matTransform, matModelview);

            int[][] tv = new int[model.vertex.length][];
            for(int i=0; i < model.vertex.length; i++)
                tv[i] = transform(model.vertex[i]);

            int[] c = {
                (int)(color[0] * 255), (int)(color[1] * 255), (int)(color[2] * 255)
            };
            for(int i=0; i < model.edge.length; i++) {
                int[][] te = {
                    tv[model.edge[i][0]], tv[model.edge[i][1]], c
                };
                if((te[0][3] + te[1][3]) < 3)
                    drawLine(te);
            }
        }
    }
    public void drawModel(Model model, double[][] color) {
        if(!mode_surface) {
            //C[t[
            matTransform = Array.copy(matProjection);
            Matrix.multiply(matTransform, matModelview);

            int[][] tv = new int[model.vertex.length][];
            for(int i=0; i < model.vertex.length; i++)
                tv[i] = transform(model.vertex[i]);

            for(int i=0; i < model.edge.length; i++) {
                int[][] te = {
                    tv[model.edge[i][0]], tv[model.edge[i][1]],
                    {(int)(color[i][0] * 255), (int)(color[i][1] * 255), (int)(color[i][2] * 255)}
                };
                if((te[0][3] + te[1][3]) < 3)
                    drawLine(te);
            }
        }
    }
    public void drawModel(Model model, Material[] material) {
        if(mode_surface) {
            if(mode_smooth) {
                if(mode_vertex) {
                    //X[XVF[fBOiގ̐ݒ肪_Ƃ̏ꍇj
                    int[][] tlv = new int[model.vertex.length][];
                    for(int i=0; i < model.vertex.length; i++)
                        tlv[i] = transform_and_lighting(model.vertex[i], model.normal_vertex[i], material[i]);

                    for(int i=0; i < model.face.length; i++) {
                        int[][] tlf = {
                            tlv[model.face[i][0]], tlv[model.face[i][1]], tlv[model.face[i][2]]
                        };
                        if((tlf[0][6] + tlf[1][6] + tlf[2][6]) < 3)
                            drawSmoothTriangle(tlf);
                    }
                } else {
                    //X[XVF[fBOiގ̐ݒ肪ʂƂ̏ꍇj
                    double[][] tlv = new double[model.vertex.length][];
                    for(int i=0; i < model.vertex.length; i++)
                        tlv[i] = transform_and_lighting2(model.vertex[i], model.normal_vertex[i]);

                    for(int i=0; i < model.face.length; i++) {
                        if((tlv[model.face[i][0]][5] + tlv[model.face[i][1]][5] + tlv[model.face[i][2]][5]) > 2)
                            continue;
                        int[][] tlf = {
                            color(tlv[model.face[i][0]], material[i]),
                            color(tlv[model.face[i][1]], material[i]),
                            color(tlv[model.face[i][2]], material[i])
                        };
                        drawSmoothTriangle(tlf);
                    }
                }
            } else {
                //tbgVF[fBO
                matTransform = Array.copy(matProjection);
                Matrix.multiply(matTransform, matModelview);

                int[][] tv = new int[model.vertex.length][];
                for(int i=0; i < model.vertex.length; i++)
                    tv[i] = transform(model.vertex[i]);

                for(int i=0; i < model.face.length; i++) {
                    int[][] tlf = {
                        tv[model.face[i][0]], tv[model.face[i][1]], tv[model.face[i][2]],
                        lighting(model.vertex[model.face[i][0]], model.vertex[model.face[i][1]],
                                model.vertex[model.face[i][2]], model.normal_face[i], material[i])
                    };
                    if((tlf[0][3] + tlf[1][3]+ tlf[2][3] + tlf[3][3]) < 3)
                        drawFlatTriangle(tlf);
                }
            }
        }
    }

    //vCx[g\bh : gXtH[CeBO - X[XVF[fBOiގ̐ݒ肪_Ƃ̏ꍇj
    private int[] transform_and_lighting(double[] vertex, double[] normal, Material material) {
        int[] tlv = new int[7];
        double[] tv = Matrix.transform3(matModelview, vertex);
        double[] tn = Vector.normalize(Vector.sub(Matrix.transform3(matModelview, normal), tv));
        double[] vl = Vector.normalize(Vector.sub(light.position, tv));
        tv = Matrix.transform3(matProjection, tv);

        if(tn[2] < 0.0 && mode_cull)
            tlv[6] = 1;
        if(tv[2] > 1.0 || tv[2] < -1.0) {
            tlv[6] = 3;
            return tlv;
        }

        double cost = Vector.dot(vl, tn);
        double cosc = 2.0*tn[2]*cost - vl[2];
        if(cost < 0.0) cost = 0.0;
        if(cosc < 0.0) cosc = 0.0;
        else cosc = pow(cosc, material.shininess);

        double[] color = new double[3];
        for(int i=0; i<3; i++) {
            color[i] =
                light.ambient[i] * material.ambient[i] +            //ambient
                cost * light.diffuse[i] * material.diffuse[i] +     //diffuse
                cosc * light.specular[i] * material.specular[i] +   //specular
                material.emission[i];                               //emission
            if(color[i] > 1.0) color[i] = 1.0;
        }

        tv = Matrix.transform3(matViewport, tv);
        tlv[0] = (int)(tv[0]);
        tlv[1] = (int)(tv[1]);
        tlv[2] = (int)(tv[2] * 0x7FFFFFFF);
        tlv[3] = (int)(color[0] * 0x7FFFFFFF);
        tlv[4] = (int)(color[1] * 0x7FFFFFFF);
        tlv[5] = (int)(color[2] * 0x7FFFFFFF);

        return tlv;
    }

    //vCx[g\bh : gXtH[CeBO - X[XVF[fBOiގ̐ݒ肪ʂƂ̏ꍇj
    private double[] transform_and_lighting2(double[] vertex, double[] normal) {
        double[] tlv = new double[6];
        double[] tv = Matrix.transform3(matModelview, vertex);
        double[] tn = Vector.normalize(Vector.sub(Matrix.transform3(matModelview, normal), tv));
        double[] vl = Vector.normalize(Vector.sub(light.position, tv));
        tv = Matrix.transform3(matProjection, tv);

        if(tn[2] < 0 && mode_cull)
            tlv[5] = 1;
        if(tv[2] > 1.0 || tv[2] < -1.0) {
            tlv[5] = 3;
            return tlv;
        }

        double cost = Vector.dot(vl, tn);
        double cosc = 2.0*tn[2]*cost - vl[2];
        if(cost < 0.0) cost = 0.0;
        if(cosc < 0.0) cosc = 0.0;

        tv = Matrix.transform3(matViewport, tv);
        tlv[0] = tv[0];
        tlv[1] = tv[1];
        tlv[2] = tv[2] * 0x7FFFFFFF;
        tlv[3] = cost;
        tlv[4] = cosc;

        return tlv;
    }
    private int[] color(double[] tlv, Material material) {
        int[] ctlv = new int[6];

        double cost = tlv[3];
        double cosc = pow(tlv[4], material.shininess);

        double[] color = new double[3];
        for(int i=0; i<3; i++) {
            color[i] =
                light.ambient[i] * material.ambient[i] +            //ambient
                cost * light.diffuse[i] * material.diffuse[i] +     //diffuse
                cosc * light.specular[i] * material.specular[i] +   //specular
                material.emission[i];                               //emission
            if(color[i] > 1.0) color[i] = 1.0;
        }

        ctlv[0] = (int)(tlv[0]);
        ctlv[1] = (int)(tlv[1]);
        ctlv[2] = (int)(tlv[2]);
        ctlv[3] = (int)(color[0] * 0x7FFFFFFF);
        ctlv[4] = (int)(color[1] * 0x7FFFFFFF);
        ctlv[5] = (int)(color[2] * 0x7FFFFFFF);

        return ctlv;
    }

    //vCx[g\bh : gXtH[ - C[t[AtbgVF[fBO
    private int[] transform(double[] vertex) {
        int[] tvi = new int[4];
        double[] tv = Matrix.transform3(matTransform, vertex);

        if(tv[2] > 1.0 || tv[2] < -1.0) {
            tvi[3] = 3;
            return tvi;
        }

        tv = Matrix.transform3(matViewport, tv);
        tvi[0] = (int)(tv[0]);
        tvi[1] = (int)(tv[1]);
        tvi[2] = (int)(tv[2] * 0x7FFFFFFF);

        return tvi;
    }

    //vCx[g\bh : CeBO - tbgVF[fBO
    private int[] lighting(double[] vertex1, double[] vertex2, double[] vertex3, double[] normal, Material material) {
        double[] cvertex = new double[3];
        for(int i=0; i<3; i++)
            cvertex[i] = (vertex1[i] + vertex2[i] + vertex3[i]) / 3.0;

        int[] lv = new int[4];
        double[] tv = Matrix.transform3(matModelview, cvertex);
        double[] tn = Vector.normalize(Vector.sub(Matrix.transform3(matModelview, normal), tv));
        double[] vl = Vector.normalize(Vector.sub(light.position, tv));

        if(tn[2] < 0 && mode_cull) {
            lv[3] = 3;
            return lv;
        }

        double cost = Vector.dot(vl, tn);
        double cosc = 2.0*tn[2]*cost - vl[2];
        if(cost < 0.0) cost = 0.0;
        if(cosc < 0.0) cosc = 0.0;
        else cosc = pow(cosc, material.shininess);

        double[] color = new double[3];
        for(int i=0; i<3; i++) {
            color[i] =
                light.ambient[i] * material.ambient[i] +            //ambient
                cost * light.diffuse[i] * material.diffuse[i] +     //diffuse
                cosc * light.specular[i] * material.specular[i] +   //specular
                material.emission[i];                               //emission
            if(color[i] > 1.0) color[i] = 1.0;
        }

        lv[0] = (int)(color[0] * 255);
        lv[1] = (int)(color[1] * 255);
        lv[2] = (int)(color[2] * 255);

        return lv;
    }

    //vCx[g\bh : Op`̕`
    private void drawSmoothTriangle(int[][] tlv) {
        int[] dx = new int[3];
        int[] dy = new int[3];
        int[] dz = new int[3];
        int[] dr = new int[3];
        int[] dg = new int[3];
        int[] db = new int[3];

        int ymin, ymax;
        int xl, xr, zl = 0, zr = 0, rl = 0, rr = 0, gl = 0, gr = 0, bl = 0, br = 0;
        int wx, wz, wr, wg, wb, difx, difz = 0, difr = 0, difg = 0, difb = 0;
        int x1, x2, z, r, g, b;
        double dw;

        for(int i=0, j=1; i<3; i++, j++) {
            if(j == 3) j = 0;
            dx[i] = tlv[j][0] - tlv[i][0];
            dy[i] = tlv[j][1] - tlv[i][1];
            dz[i] = tlv[j][2] - tlv[i][2];
            dr[i] = tlv[j][3] - tlv[i][3];
            dg[i] = tlv[j][4] - tlv[i][4];
            db[i] = tlv[j][5] - tlv[i][5];
        }

        ymin = tlv[0][1];
        ymax = tlv[0][1];
        for(int i=1; i<3; i++) {
            if(tlv[i][1] < ymin) ymin = tlv[i][1];
            else if(tlv[i][1] > ymax) ymax = tlv[i][1];
        }
        if(ymin < 0) ymin = 0;
        if(ymax > height) ymax = height;

        for(int y = ymin; y < ymax; y++) {
            xl = 0x7FFFFFFF;
            xr = 0x80000000;
            for(int i=0; i<3; i++) {
                if(dy[i] == 0) continue;
                dw = (double)(y - tlv[i][1]) / dy[i];
                if((dw >= 0.0) && (dw <= 1.0)) {
                    wx = (int)(dw * dx[i]) + tlv[i][0];
                    wz = (int)(dw * dz[i]) + tlv[i][2];
                    wr = (int)(dw * dr[i]) + tlv[i][3];
                    wg = (int)(dw * dg[i]) + tlv[i][4];
                    wb = (int)(dw * db[i]) + tlv[i][5];
                    if(wx < xl) {
                        xl = wx;
                        zl = wz;
                        rl = wr;
                        gl = wg;
                        bl = wb;
                    }
                    if(wx > xr) {
                        xr = wx;
                        zr = wz;
                        rr = wr;
                        gr = wg;
                        br = wb;
                    }
                }
            }

            difx = xr - xl;
            if(difx != 0) {
                 difz = (zr - zl) / difx;
                 difr = (rr - rl) / difx;
                 difg = (gr - gl) / difx;
                 difb = (br - bl) / difx;
            }

            x1 = xl;
            x2 = xr;
            if(x1 < 0) x1 = 0;
            if(x2 > (width - 1)) x2 = width - 1;

            z = zl + (x1 - xl) * difz;
            r = rl + (x1 - xl) * difr;
            g = gl + (x1 - xl) * difg;
            b = bl + (x1 - xl) * difb;
            for(int x = x1, n = x1 + y * width; x <= x2; x++, n++) {
                if(z > zbuf[n]) {
                    pbuf[n] = (255<<24) | ((r>>23)<<16) | ((g>>23)<<8) | (b>>23);
                    zbuf[n] = z;
                }
                z += difz;
                r += difr;
                g += difg;
                b += difb;
            }
        }
    }
    private void drawFlatTriangle(int[][] tlv) {
        int[] dx = new int[3];
        int[] dy = new int[3];
        int[] dz = new int[3];

        int ymin, ymax;
        int xl, xr, zl = 0, zr = 0;
        int wx, wz, difx, difz = 0;
        int x1, x2, z, c;
        double dw;

        c = (255<<24) | (tlv[3][0]<<16) | (tlv[3][1]<<8) | tlv[3][2];

        for(int i=0, j=1; i<3; i++, j++) {
            if(j == 3) j = 0;
            dx[i] = tlv[j][0] - tlv[i][0];
            dy[i] = tlv[j][1] - tlv[i][1];
            dz[i] = tlv[j][2] - tlv[i][2];
        }

        ymin = tlv[0][1];
        ymax = tlv[0][1];
        for(int i=1; i<3; i++) {
            if(tlv[i][1] < ymin) ymin = tlv[i][1];
            else if(tlv[i][1] > ymax) ymax = tlv[i][1];
        }
        if(ymin < 0) ymin = 0;
        if(ymax > height) ymax = height;

        for(int y = ymin; y < ymax; y++) {
            xl = 0x7FFFFFFF;
            xr = 0x80000000;
            for(int i=0; i<3; i++) {
                if(dy[i] == 0) continue;
                dw = (double)(y - tlv[i][1]) / dy[i];
                if((dw >= 0.0) && (dw <= 1.0)) {
                    wx = (int)(dw * dx[i]) + tlv[i][0];
                    wz = (int)(dw * dz[i]) + tlv[i][2];
                    if(wx < xl) {
                        xl = wx;
                        zl = wz;
                    }
                    if(wx > xr) {
                        xr = wx;
                        zr = wz;
                    }
                }
            }

            difx = xr - xl;
            if(difx != 0)
                 difz = (zr - zl) / difx;

            x1 = xl;
            x2 = xr;
            if(x1 < 0) x1 = 0;
            if(x2 > (width - 1)) x2 = width - 1;

            z = zl + (x1 - xl) * difz;
            for(int x = x1, n = x1 + y * width; x <= x2; x++, n++) {
                if(z > zbuf[n]) {
                    pbuf[n] = c;
                    zbuf[n] = z;
                }
                z += difz;
            }
        }
    }

    //vCx[g\bh : ̕`
    private void drawLine(int[][] tv) {
        int x, y, z, dx, dy, dz, dx2, dy2, ex, ey, wx, wy, c, n;

        x = tv[0][0];
        y = tv[0][1];
        z = tv[0][2];
        dx = tv[1][0] - tv[0][0];
        dy = tv[1][1] - tv[0][1];
        dz = tv[1][2] - tv[0][2];
        dx2 = 2*Math.abs(dx);
        dy2 = 2*Math.abs(dy);
        wx = Math.abs(dx) + 1;
        wy = Math.abs(dy) + 1;

        c = (255<<24) | (tv[2][0]<<16) | (tv[2][1]<<8) | tv[2][2];

        if(wx > wy) {
            ey = -Math.abs(dx);
            dz /= wx;
            for(int i=0; i<wx; i++) {
                if(x >= 0 && x < width && y >= 0 && y < height){
                    n = x + y*width;
                    if(z > zbuf[n]) {
                        pbuf[n] = c;
                        zbuf[n] = z;
                    }
                }
                x += addDir(dx);
                ey += dy2;
                z += dz;
                if(ey >= 0) {
                    y += addDir(dy);
                    ey -= dx2;
                }
            }
        } else {
            ex = -Math.abs(dy);
            dz /= wy;
            for(int i=0; i<wy; i++) {
                if(x >= 0 && x < width && y >= 0 && y < height){
                    n = x + y*width;
                    if(z > zbuf[n]) {
                        pbuf[n] = c;
                        zbuf[n] = z;
                    }
                }
                y += addDir(dy);
                ex += dx2;
                z += dz;
                if(ex >= 0) {
                    x += addDir(dx);
                    ex -= dy2;
                }
            }
        }
    }
    private int addDir(int dxy) {
        if(dxy < 0) return -1;
        else if(dxy == 0) return 0;
        else return 1;
    }

    //ab
    private double pow(double a, int b) {
        double p = 1.0;
        for(; b>0; b--) p *= a;
        return p;
    }
}